import numpy as np
import matplotlib.pyplot as plt
import time

def HammerHollingsworth(T, h, y0, f, df, maxIter = 12, tol = 1e-14, debug=False):
    n = y0.shape[0]

    b = np.array([0.5,0.5])
    c = np.array([1/2-1/6*np.sqrt(3), 1/2+1/6*np.sqrt(3)])
    a = np.array([[1/4, 1/4-1/6*np.sqrt(3)], [1/4+1/6*np.sqrt(3), 1/4]])

    def G(k, x, y, f):
        g = np.zeros(2*n)
        g[:n] = k[0] - y - h*a[0,0]*f(x+h*c[0], k[0]) - h*a[0,1]*f(x+h*c[1], k[1])
        g[n:] = k[1] - y - h*a[1,0]*f(x+h*c[0], k[0]) - h*a[1,1]*f(x+h*c[1], k[1])
        return g

    def dG(k, x, y, df):
        A=np.zeros((2*n,2*n))
        A[:n,:n] = np.eye(n) - h*a[0,0]*df(x+h*c[0], k[0])
        A[:n,n:] = - h*a[0,1]*df(x+h*c[0], k[1])
        A[n:,:n] = - h*a[1,0]*df(x+h*c[1], k[0])
        A[n:,n:] = np.eye(n) - h * a[1, 1] * df(x + h * c[1], k[1])
        return A

    k = np.array([y0, y0])
    y = [y0]
    ti = [0]
    t = 0
    start = time.time()
    while t < T:
        iter = 0
        m=G(k, t, y[-1], f)
        while np.linalg.norm(m,np.inf) > tol and iter < maxIter:
            A = dG(k, t, y[-1], df)
            s = np.linalg.solve(A, m)
            if debug:
                print(A,s)
            k[0] -= s[:n]
            k[1] -= s[n:]
            iter += 1
            m = G(k,t,y[-1], f)

        if np.linalg.norm(m) > tol and iter == maxIter:
            print('No convergence in Newton Iteration')
            break
            # raise Exception('No convergence in Newton Iteration')

        y.append(y[-1] + h*b[0]*f(t+h*c[0], k[0]) + h*b[1]*f(t+h*c[1], k[1]))
        t += h
        ti.append(t)
    end = time.time()

    y = np.array(y)
    ti = np.array(ti)

    #print('cpu-time:',end-start)
    return (ti,y)

def classicRungeKutta(T, h, y0, f):
    y = [y0]
    ti = [0]
    t = 0
    start = time.time()
    while t < T:
        r1 = f(t,y[-1])
        r2 = f(t+h/2,y[-1]+h/2*r1)
        r3 = f(t+h/2,y[-1]+h/2*r2)
        r4 = f(t+h,y[-1]+h*r3)

        y.append(y[-1] + h/6*(r1+2*r2+2*r3+r4))
        t += h
        ti.append(t)
    end = time.time()

    y = np.array(y)
    ti = np.array(ti)

    #print('cpu-time:',end-start)
    return (ti,y)

def Runge(T, h, y0, f):
    y = [y0]
    ti = [0]
    t = 0
    start = time.time()
    while t < T:
        r1 = f(t, y[-1])
        r2 = f(t+h/2, y[-1]+h/2*r1)

        y.append(y[-1] + h*r2)
        t += h
        ti.append(t)
    end = time.time()

    y = np.array(y)
    ti = np.array(ti)

    #print('cpu-time:',end-start)
    return (ti,y)

def Heun(T, h, y0, f):
    y = [y0]
    ti = [0]
    t = 0
    start = time.time()
    while t < T:
        k1 = y[-1]
        k2 = y[-1] + h * f(t,k1)

        y.append(y[-1] + h/2*(f(t,k1)+f(t+h,k2)))
        t += h
        ti.append(t)
    end = time.time()

    y = np.array(y)
    ti = np.array(ti)

    #print('cpu-time:',end-start)
    return (ti,y)

def explicitEuler(T, h, y0, f):
    y = [y0]
    ti = [0]
    t = 0
    start = time.time()
    while t < T:
        y.append(y[-1] + h*f(t,y[-1]))
        t += h
        ti.append(t)
    end = time.time()

    y = np.array(y)
    ti = np.array(ti)

    #print('cpu-time:',end-start)
    return (ti,y)


if __name__ == "__main__":
    # Modell Beispiel
    def f(x,y):
        return np.array([y[1],-y[0]])
    
    def df(x,y):
        return np.array([[0.,1.],[-1.,0.]])
    
    # Mathematisches Pendel
    def fpendel(x,y):
        return np.array([y[1],-np.sin(y[0])])
    
    def dfpendel(x,y):
        return np.array([[0.,1.],[-np.cos(y[0]),0.]])
    
    y0 = np.array([3.14125,0])
    
    (t,y) = HammerHollingsworth(8000, 1, y0, f, df)
    (t,y) = HammerHollingsworth(400, .1, y0, fpendel, dfpendel)
    
    plt.plot(t,y)
    plt.grid()
    plt.show()
    
    plt.plot(y[:,0],y[:,1],'-')
    plt.grid()
    plt.show()
